from paver.defaults import Bunch

options(
    setup=Bunch(
        name="metamake",
        packages=['metamake'],
        version="1.0.6",
        license="BSD",
        author="Matt Pizzimenti",
        author_email="mjpizz+metamake@gmail.com",
        url="http://pypi.python.org/pypi/metamake/",
        description="Metamake is a dead-simple task-based automation tool written in Python.",
        long_description='''\
Description
===========

Metamake **is a simple way to define common tasks** and execute those tasks by
name, similar to Rake.  Metamake **is not** a dependency-tracking build tool
like Make, ant, qmake, SCons, Visual Studio, or XCode.  Metamake is used
**with** these build tools to orchestrate complex builds that work in a
cross-platform fashion.

Usage
=====
Using Metamake is as easy as creating a 'Makefile.py' in your project
directory::

    from metamake import task, shell, path

    @task
    def build()
        """builds the widget"""
        shell("qmake proj.pro -o Makefile.proj && make -f Makefile.proj")
        path("src/headers").copytree("dist/include")

On the commandline, you can then type ``metamake ls`` to see a listing
of all Metamake tasks defined in your Makefile.py, with their docstrings
helpfully listed to describe the purpose of that task.

Advanced Features
=================

Backwards-compatibility with Make
---------------------------------
To make things easier on newcomers to Metamake, a "Makefile" can be
created in your working directory that contains a bootstrapped version of
Metamake inside.  This allows anybody to build your project without needing
Metamake to be installed.  With the bootstrapped Makefile, you can type
``make <args>`` to achieve the same effect as ``metamake <args>``.  Whenever you
update Metamake on your system, these bootstrapped Makefiles will be updated
automatically next time you execute Metamake for that project.  You should 
commit these Makefiles to your repository so that other people can check out
your project and build it without installing Metamake.  To create a Makefile
bootstrap::

    from metamake import task, shell, bootstrap
    
    bootstrap("Makefile")
    
    # ...

What if you already have a Makefile that you are using for other purposes?
That's easy to solve, simply by specifying a different filename::
    
    from metamake import task, shell, bootstrap
    
    bootstrap("Makefile.meta")
    
    # ...
    
Easy Commandline Flag Definition
--------------------------------
Metamake allows you to define commandline flags that can be passed
into your build process.  These flags will work regardless of whether you use
the 'metamake' tool or the bootstrapped Makefile::

    from metamake import task, Flag

    Flag("cleanfirst").explain("set this flag to 'true' to do a clean build")
    
    if Flag("cleanfirst").given:
        print "flag was given"
    
    if Flag("cleanfirst").value == "true":
        print "flag value was True"

When you execute ``metamake ls`` on the commandline, you will see these flags
listed underneath all of the task definitions, with the explanation that
you provided as documentation.

From the first example, asking if the value is "true" or "false" is a bit unweildy
for a boolean value.  This gets even worse for flags that take a restricted set of
values, for example a range of integers.  The ``explain`` method can take a few
more parameters to help you out here::

    from metamake import task, Flag

    Flag("debuglevel").explain("set the debug level", convert=int, allow=[1,2,3,4])
    
    if Flag("debuglevel").value == 3:
        print "flag was converted to the integer 3"

You can also grab the Flag instance instead of always referring to the
string name::

    from metamake import task, Flag
    
    cleanflag = Flag("cleanfirst")
    debugflag = Flag("debuglevel")
    
    if cleanflag.given and debugflag.value == 3:
        print "we just checked the flag instances"
    

Readable, Cross-platform Filesystem Manipulation
------------------------------------------------
Jason Orendorff's excellent `path.py <http://pypi.python.org/pypi/path.py/>`_
library unifies all of the cross-platform Python filesystem manipulations
under a single object called **path**::

    from metamake import task, path

    @task
    def build():
        """builds the widget"""
        path("dist/include").makedirs()
        path("src/widget").copytree("dist/include/widget")
        for header_file in path("src/gadget").listdir("*.h"):
            header_file.copyfile("dist/include/gadget/%s" % header_file.basename())

Metamake extends Jason's library by providing console logging for file operations.
This makes it easy to see the manipulations that are happening to your filesystem
on the commandline.

External Dependency Tracking (alpha-quality)
--------------------------------------------
Metamake also supports the concept of Dependencies.  This allows you to specify
external dependencies for your project, without setting up messy hierarchical
builds using svn:externals or similar tools.  Every Dependency has three
methods:  ``update``, ``build``, and ``clean``::

    from metamake import task, shell, Dependency

    gadgetlib = Dependency(
        url = "svn://myrepo.com/gadgetlib",
        path = "~/.gadgetlib",
        build_cmd = "make",
        clean_cmd = "make clean",
    )

    @task
    def build()
        """builds the widget"""
        gadgetlib.update()
        gadgetlib.build()
        shell("qmake")
''',
        entry_points='''
[console_scripts]
mm = metamake:metamake_bin
metamake = metamake:metamake_bin
''',
        zip_safe = False,
        minilib=Bunch()
    )
)

@task
@needs(['generate_setup', 'minilib', 'setuptools.command.sdist'])
def sdist():
    """Overrides sdist to make sure that our setup.py is generated."""
    pass



from logging import getLogger
from os import system

def shell(cmd, quiet=False, allow_failure = False):
    """runs a shell command, echoing output and raising
    a SystemExit if the command fails"""
    logger = getLogger("metamake.shell")
    if not quiet:
        logger.info(cmd)
    retcode = system(cmd)
    if retcode is not 0 and not allow_failure:
        raise SystemExit(retcode)


# TODO: bootstrap needs to check for the metamake-bootstrap header
from __future__ import generators
import sys
from logging import getLogger, StreamHandler, INFO
from metamake._path import loggable_path as path
from metamake._bootstrap import serialize_metamake_to_makefile
from metamake._flags import Flag
from metamake._dependencies import Dependency
from metamake._shell import shell

__all__ = [
    
    "task",
    "shell"
    "path",
    "Flag",
    "Dependency",
    
    "list_tasks",
    "run_task",
    "run_default_task",
]

_DEFAULT_STDERR_LOGS = ["metamake.filesystem", "metamake.shell"]
_all_tasks = {}
_ordered_task_keys = []
_first_task_name = None
_default_task_name = None
_did_setup_stderr_logging = False
_bootstrap_filename = None

def bootstrap(bootstrap_filename):
    """call this function to set a bootstrap filename for
    serializing metamake to a Makefile for backwards-compat"""
    global _bootstrap_filename
    _bootstrap_filename = bootstrap_filename

def task(method=None, default=False):
    """this decorator should be used to mark Python functions
    in your Makefile.py as tasks that can be executed on
    the commandline.  You can specify 'default = True' if
    you want that task to be the default"""
    def decorate(method, default):

        global _first_task_name
        global _default_task_name
        
        if _first_task_name == None:
            _first_task_name = method.__name__
        
        if default == True:
            _default_task_name = method.__name__
        
        _all_tasks[method.__name__] = method
        _ordered_task_keys.append(method.__name__)
        return method
        
    if method == None:
        def subdecorator(method):
            return decorate(method, default)
        return subdecorator
    else:
        return decorate(method, default)

def list_tasks():
    """lists all tasks that have been defined in your
    Makefile.py file"""
    _ensure_stderr_logs_are_active()
    if _bootstrap_filename:
        serialize_metamake_to_makefile(_bootstrap_filename)
    Flag("noupdate").explain("set to 'true' if you want to disable automatic Dependency version control updates")
    print
    if len(_all_tasks) is 0:
        print "(no tasks defined)"
    else:
        for task_name in _ordered_task_keys:
            task = _all_tasks[task_name]
            format_string = "%%(name)-%ss # %%(optional_default_doc)s%%(doc)s" % max([len(name)+1 for name in _all_tasks.keys()])
            print format_string % dict(
                name = task.__name__,
                doc = task.__doc__,
                optional_default_doc = {_default_task_name: "(default) "}.get(task_name, ""),
            )
    print
    for flagname, explanation in Flag._flag_explanation_lookup.iteritems():
        print "[%(flagname)s=?] %(explanation)s" % locals()
    print

def run_task(task_name):
    """runs a task (by name)"""
    _ensure_stderr_logs_are_active()
    if _bootstrap_filename:
        serialize_metamake_to_makefile(_bootstrap_filename)
    global _all_tasks
    if len(_all_tasks) is 0:
        print >> sys.stderr, "*** no tasks were defined (did you forget to define tasks in the 'Makefile.py' in your working directory?)"
        raise SystemExit(1)
    if task_name in _all_tasks:
        _all_tasks[task_name]()
    else:
        print >> sys.stderr, "*** there is no task with the name '%s'" % task_name
        raise SystemExit(1)

def run_default_task():
    """runs the task marked as default (using '@task(default=True)')"""
    global _first_task_name
    global _default_task_name
    
    if _default_task_name:
        run_task(_default_task_name)
    
    else:
        run_task(_first_task_name)

def _ensure_stderr_logs_are_active():
    """helper method to ensure that file system actions
    are written to stderr"""
    global _did_setup_stderr_logging
    if not _did_setup_stderr_logging:
        stderr_logging_handler = StreamHandler(sys.stderr)
        for logger_name in _DEFAULT_STDERR_LOGS:
            logger = getLogger(logger_name)
            logger.setLevel(INFO)
            logger.addHandler(stderr_logging_handler)
        _did_setup_stderr_logging = True

def metamake_bin():
    """the actual metamake commandline tool (main function)"""
    
    import sys
    
    # allow help from the commandline
    if "--help" in sys.argv:
        print >> sys.stderr, "latest metamake documentation can be found at http://pypi.python.org/pypi/metamake/"
        raise SystemExit(0)

    # the user didn't want help, they are actually trying to exec metamake
    import os
    sys.path.insert(0, os.getcwd())
    import metamake
    
    try:
        import Makefile
    except ImportError, e:
        
        errmsg = str(e)
        if "Makefile" in errmsg:
            # this was probably unable to find Makefile.py
            print >> sys.stderr, "*** Makefile.py could not be loaded.  It appears that you have not defined one."
        else:
            # this was probably due to an import from within Makefile.py
            print >> sys.stderr, "*** Makefile.py could not be loaded.  It appears that one of your imports is failing: '%s'." % errmsg
        
        # fail with nonzero error code
        raise SystemExit(1)
    
    # any commandline argument without a '=' in it is a task that we want to run
    tasknames = [arg for arg in sys.argv[1:] if "=" not in arg]
    
    # decide what task(s) to run
    if len(tasknames) is 0:
        
        # no tasknames means we should run the default one
        run_default_task()
        
    elif "ls" in tasknames:
        
        # the 'ls' task is a special task that lists all other tasks
        list_tasks()
    else:
        
        # otherwise, iterate through all task names and run them
        for taskname in tasknames:
            run_task(taskname)



import sys
from os import environ

def _get_args_from_commandline():
    """helper method that gets the commandline arguments
    regardless of whether metamake was run as a standalone
    app or through a Makefile bootstrap"""
    
    if "MAKEFLAGS" in environ:
        
        # we were run via 'make', use the MAKEFLAGS
        # environment variable to see our flags
        makeflags = environ["MAKEFLAGS"].split()
        return [arg for arg in makeflags if "=" in arg]
        
    else:
        
        # we were run standalone, use any commandline
        # arguments that have an '=' in them
        return [arg for arg in sys.argv if "=" in arg]


class Flag(object):
    """represents a flag that can be given on the commandline
    to affect a build"""
    
    _flag_value_lookup = None
    _flag_explanation_lookup = {}
    _flag_type_lookup = {}
    _flag_allowablevals_lookup = {}
    
    def _get_value(self):
        if not self.given:
            return None
        else:
            value = Flag._flag_value_lookup[self.__flag_name]
            type_ = Flag._flag_type_lookup.get(self.__flag_name, None)
            allowable_values = Flag._flag_allowablevals_lookup.get(self.__flag_name, None)
            def _raise_value_error():
                raise ValueError("invalid value %s for flag '%s')" % (repr(value), self.__flag_name))
            if type_ is not None:
                if issubclass(type_, bool):
                    # we treat boolean differently to ensure that it maps properly
                    if value.lower() in ["true", "1", "t", "yes", "y", "on"]:
                        value = True
                    elif value.lower() in ["false", "0", "f", "no", "n", "off"]:
                        value = False
                    else:
                        _raise_value_error()
                else:
                    # all other values can be converted properly
                    value = type_(value)
            if allowable_values is not None:
                if value not in allowable_values:
                    _raise_value_error()
            return value
    
    given = property(lambda self: bool(self.__flag_name in Flag._flag_value_lookup))
    """returns true if the Flag has been given on the commandline (regardless of value)"""
    
    value = property(_get_value)
    """returns the actual value of the Flag (if given)"""
    
    explanation = property(lambda self: Flag._flag_explanation_lookup.get(self.__flag_name, None))
    """return a human-readable explanation of this Flag's affect"""
    
    def explain(self, explanation, convert=None, allow=None):
        """this method is used to explain a flag in plain
        language so that the 'ls' task will display it"""

        if self.__flag_name in Flag._flag_explanation_lookup.keys():
        
            # this Flag was already explained once, don't allow re-explanation
            print >> sys.stderr, "*** the Flag '%(flagname)s' is already being used for '%(explanation)s" % dict(
                flagname = self.__flag_name,
                explanation = Flag._flag_explanation_lookup[self.__flag_name],
            )
            raise SystemExit(1)
        
        else:
            
            # no conflicts, set the explanation
            Flag._flag_explanation_lookup[self.__flag_name] = explanation
            if convert is not None:
                Flag._flag_type_lookup[self.__flag_name] = convert
            if allow is not None:
                Flag._flag_allowablevals_lookup[self.__flag_name] = allow

        
        # return the Flag object to allow chaining
        return self

    def __init__(self, flag_name):
        
        # lazy-parse the makeflags
        if Flag._flag_value_lookup == None:
            Flag._flag_value_lookup = {}
            for key_and_value in _get_args_from_commandline():
                key,value = key_and_value.split("=", 1)
                Flag._flag_value_lookup[key] = value
        
        # keep information about this flag
        self.__flag_name = flag_name



import platform
from metamake._flags import Flag
from metamake._path import loggable_path
from metamake._shell import shell

def _make(target, path):
    """helper method for running make (tries to use nmake on windows)"""
    from metamake import shell
    if platform.system() == "Windows":
        makebin = "nmake"
    else:
        makebin = "make"
    shell("cd %(path)s && %(makebin)s %(target)s" % dict(
        path = path,
        makebin = makebin,
        target = target,
    ))


class Dependency(object):
    """represents an external dependency for your project.
    Given a URL, this object will figure out the type of version
    control and pull down the dependency into the requested path"""
    
    def __init__(self, url, path, build_cmd=None, clean_cmd=None):

        self.__url = url
        self.__path = path

        if build_cmd:
            self.__build_cmd = build_cmd
        else:
            self.__build_cmd = lambda: _make("", self.__path)

        if clean_cmd:
            self.__clean_cmd = clean_cmd
        else:
            self.__clean_cmd = lambda: _make("clean", self.__path)
    
    def update(self):
        """updates this dependency from its version control"""
        
        if not Flag("noupdate").given or Flag("noupdate").value.lower() is not "true":
            
            # the 'noupdate' flag was not given, do an svn update
            if not loggable_path(self.__path).exists():
                
                # checkout did not exist yet, check it out first
                shell("svn co %(url)s %(path)s" % dict(
                    url = self.__url,
                    path = self.__path,
                ))
                
            else:
                
                # checkout existed, do an svn update
                shell("cd %(path)s && svn revert Makefile && svn up" % dict(
                    path = self.__path
                ))
        else:
            
            # the 'noupdate' flag was given on the commandline, don't do subversion updates
            print >> sys.stderr, "*** skipping Dependency Subversion updates"
    
    def build(self):
        """builds this dependency.  This will do an initial checkout
        if the project has not been checked out yet"""
        # if we haven't done a checkout yet, do it
        if not loggable_path(self.__path).exists():
            self.update()
        
        if isinstance(self.__build_cmd, basestring):
            
            # the command was a string, exec it on the commandline
            shell("cd %(path)s && %(build_cmd)s" % dict(
                path = self.__path,
                build_cmd = self.__build_cmd,
            ))
            
        else:
            
            # the command was a callback, use it
            self.__build_cmd()
    
    def clean(self):
        """cleans the build for the dependency"""
        
        # if we haven't done a checkout yet, do it
        if not loggable_path(self.__path).exists():
            self.update()
            
        if isinstance(self.__clean_cmd, basestring):

            # the command was a string, exec it on the commandline
            shell("cd %(path)s && %(clean_cmd)s" % dict(
                path = self.__path,
                clean_cmd = self.__clean_cmd,
            ))
            
        else:
            
            # the command was a callback, use it
            self.__clean_cmd()
